package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.AbstractObject2ObjectFunction;
import it.unimi.di.law.vector.Vector;



/**
 * An abstract class implementing the basic functions of a featurizer.
 * 
 * <p>Concrete subclasses should just implement {@link #get(Object)}, returning a vector which
 * represents the features.
 * 
 * <p>Featurizers are used to combine archive data with term/count document information to build
 * (weighted) scores for terms. For example, if term <code>t</code> is in the document, then the
 * <i>t</i>-th coordinate of the vector contains its score.
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */
public abstract class AbstractFeaturizer<T extends DocumentSummary, V extends Vector> extends AbstractObject2ObjectFunction<T, V> {
	private static final long serialVersionUID = 1L;

	/** Number of documents in the archive. */
	protected final int numberOfDocuments;

	/** Archive. */
	protected final Archive<? extends T> archive;

	public AbstractFeaturizer( final Archive<? extends T> archive ) {
		this.archive = archive;
		this.numberOfDocuments = archive.numberOfDocuments();
	}

	/**
	 * Returns whether the argument is an instance of {@link DocumentSummary}.
	 * 
	 * @param o an object.
	 * @return true the argument is an instance of {@link DocumentSummary}.
	 */
	public boolean containsKey( Object o ) {
		return o instanceof DocumentSummary;
	}

	/** Returns the number of documents in the archive. */
	public int size() {
		return numberOfDocuments;
	}

}
