package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.lang.FlyweightPrototype;

import java.io.Closeable;
import java.io.IOException;

/**
 * An immutable archive of {@linkplain DocumentSummary document summaries}.
 * 
 * <p>Archives provide sequential access using an iterator, and random optional access by
 * {@linkplain #getDocumentById(int) document id} or by
 * {@linkplain #getDocumentByIndex(int) document index} (you can check for random-access
 * availability using {@link #hasRandomAccess()}). Some global data about the archive is also
 * available.
 * 
 * @see ArchiveLoader
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */
public interface Archive<T extends DocumentSummary> extends Closeable, Iterable<T>, FlyweightPrototype<Archive<T>> {

	/**
	 * Returns a document given its id (optional operation).
	 * 
	 * @param id a document id.
	 * @return the document with given id, or <code>null</code> if no such document exists.
	 */
	public T getDocumentById( int id ) throws IOException;

	/**
	 * Returns a document by index (position in the archive) (optional operation).
	 * 
	 * @param idx the document index.
	 */
	public T getDocumentByIndex( int idx ) throws IOException;

	/**
	 * Returns whether the archive supports random access, that is, {@link #getDocumentById(int)}
	 * and {@link #getDocumentByIndex(int)}.
	 * 
	 * @return whether the archive supports random access.
	 */
	public boolean hasRandomAccess();

	/**
	 * Returns the number of documents in the archive.
	 * 
	 * @return the number of documents in the archive.
	 */
	public int numberOfDocuments();

	/**
	 * Returns the number of terms in the archive.
	 * 
	 * @return the number of terms in the archive.
	 */
	public int numberOfTerms();

	/**
	 * Returns the number of words in the collection (i.e., the sum of the lengths of all
	 * documents).
	 * 
	 * @return the number of words in the collection.
	 */
	public long numberOfWords();

	/**
	 * Return the frequency of a given term.
	 * 
	 * @param term a term number.
	 * @return the frequency of the given term.
	 */
	public int frequency( int term );

}
