package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import java.io.Closeable;
import java.io.IOException;




/**
 * An abstract archive writer.
 * 
 * <p>Concrete implementations of this class must implement {@link #append(DocumentSummary)},
 * which appends a summary to the archive, and {@link Closeable#close()}, which finalizes the
 * archive. It is strongly suggested that the {@link ArchiveLoader} properties convention is
 * followed.
 * 
 * <p>Users of concrete implementations must create an instance,
 * {@linkplain #append(DocumentSummary) append} the relevant summaries, and
 * {@linkplain Closeable#close() close} the archive.
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */

public abstract class ArchiveWriter<T extends DocumentSummary> implements Closeable {

	/**
	 * Appends a new document summary to the archive.
	 * 
	 * @param summary a document summary.
	 * 
	 */
	public abstract void append( T summary ) throws IOException;

	/**
	 * Appends an entire collection to the archive.
	 * 
	 * <p>This implementation repeatedly calls {@link #append(DocumentSummary)}.
	 */
	public void appendAll( final Iterable<T> collection ) throws IOException {
		for ( T d : collection )
			append( d );
	}
}
