package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


import java.net.URI;
import java.util.StringTokenizer;

import cern.colt.GenericSorting;
import cern.colt.Swapper;
import cern.colt.function.IntComparator;




/**
 * A simple, array-based implementation of {@link DocumentSummary}.
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */

public class ArrayDocumentSummary implements DocumentSummary {
	/** The terms in this document, in increasing order. */
	public final int term[];

	/**
	 * An array, parallel to {@link #term}, containing the counts (a.k.a. within-document
	 * frequencies) of the respective terms. It is <code>null</code> if counts are not present.
	 */
	public final int count[];

	/** The document id. */
	protected final int id;

	/** The document URI, if any, or <code>null</code>. */
	protected final URI uri;

	/** The length of the underlying document in words, or -1 if document lengths are not present. */
	protected final int docLength;

	/** Number of terms in the summary. */
	protected final int numberOfTerms;

	/** Whether {@link #term} is sorted. */
	protected boolean sorted;

	/**
	 * Creates an unsorted array-based document summary using given arrays.
	 * 
	 * @param terms the terms in the summary (in increasing order).
	 * @param counts an array parallel to <code>terms</code> specifying the counts.
	 * @param id the id of the document.
	 * @param uri a URI for the document, or <code>null</code>.
	 * @param wordLength the length in words of the underlying document.
	 */
	public ArrayDocumentSummary( int[] terms, int[] counts, int id, URI uri, int wordLength ) {
		this( terms, counts, terms.length, id, uri, wordLength );
	}

	/**
	 * Creates an array-based document summary using given arrays.
	 * 
	 * @param terms the terms in the summary (in increasing order).
	 * @param counts an array parallel to <code>terms</code> specifying the counts.
	 * @param id the id of the document.
	 * @param uri a URI for the document, or <code>null</code>.
	 * @param wordLength the length in words of the underlying document.
	 * @param sorted if true, <code>terms</code> is sorted (which makes it possible to use binary
	 * search to locate a term).
	 */
	public ArrayDocumentSummary( int[] terms, int[] counts, int id, URI uri, int wordLength, boolean sorted ) {
		this( terms, counts, terms.length, id, uri, wordLength, sorted );
	}

	/**
	 * Creates an array-based document summary using part of given arrays.
	 * 
	 * @param terms the terms in the summary (in increasing order).
	 * @param counts an array parallel to <code>terms</code> specifying the counts.
	 * @param id the id of the document.
	 * @param uri a URI for the document, or <code>null</code>.
	 * @param wordLength the length in words of the underlying document.
	 */
	public ArrayDocumentSummary( int[] terms, int[] counts, int numTerms, int id, URI uri, int wordLength, boolean sorted ) {
		this.term = terms;
		this.count = counts;
		this.id = id;
		this.uri = uri;
		this.docLength = wordLength;
		this.numberOfTerms = numTerms;
		this.sorted = sorted;
	}

	/**
	 * Creates an unsorted array-based document summary using part of given arrays.
	 * 
	 * @param terms the terms in the summary (in increasing order).
	 * @param counts an array parallel to <code>terms</code> specifying the counts.
	 * @param id the id of the document.
	 * @param uri a URI for the document, or <code>null</code>.
	 * @param wordLength the length in words of the underlying document.
	 */
	public ArrayDocumentSummary( int[] terms, int[] counts, int numTerms, int id, URI uri, int wordLength ) {
		this( terms, counts, numTerms, id, uri, wordLength, false );
	}

	/**
	 * * Creates a new array-based document summary by parsing ASCII text.
	 * 
	 * <p>This method is mainly useful for testing and debugging purposes.
	 * 
	 * @param line the ASCII text containing the description of the document summary.
	 * @param uri an optional URI that will be associated to the returned summary.
	 * @return a document summary, as specified by the ASCII text in <code>line</code>.
	 */
	public static ArrayDocumentSummary parse( String line, URI uri ) {
		// TODO: Use StreamTokenizer (faster).
		final StringTokenizer st = new StringTokenizer( line );

		int id = Integer.parseInt( st.nextToken() );

		final int wordLen = Integer.parseInt( st.nextToken() );
		final int numberOfTerms = Integer.parseInt( st.nextToken() );


		int[] terms = new int[ numberOfTerms ];
		int[] counts = new int[ numberOfTerms ];

		for ( int i = 0; i < numberOfTerms; i++ ) {
			String[] array = st.nextToken().split( ":" );

			terms[ i ] = Integer.parseInt( array[ 0 ] );
			counts[ i ] = Integer.parseInt( array[ 1 ] );
		}

		return new ArrayDocumentSummary( terms, counts, id, uri, wordLen );
	}

	public int term( int i ) {
		return term[ i ];
	}

	public int count( int i ) {
		if ( count == null ) return -1;
		return count[ i ];
	}

	public boolean sorted() {
		return sorted;
	}

	public ArrayDocumentSummary sort() {
		// Sort terms back.
		if ( sorted ) return this;
		sorted = true;
		GenericSorting.mergeSort( 0, numberOfTerms, new IntComparator() {
			public int compare( int x, int y ) {
				return term[ x ] - term[ y ];
			}

		}, new Swapper() {
			public void swap( int x, int y ) {
				int tmp = term[ x ];
				term[ x ] = term[ y ];
				term[ y ] = tmp;
				if ( count != null ) {
					tmp = count[ x ];
					count[ x ] = count[ y ];
					count[ y ] = tmp;
				}
			}
		} );

		return this;
	}


	/** This implementation is compatible for reading via {@link #parse(String, URI)}. */
	@Override
	public String toString() {
		final StringBuilder s = new StringBuilder();

		s.append( id );
		if ( docLength != -1 ) {
			s.append( " " );
			s.append( docLength );
		}
		s.append( " " );
		s.append( term.length );
		for ( int i = 0; i < numberOfTerms; i++ ) {
			s.append( " " );
			s.append( term[ i ] );
			if ( count != null ) {
				s.append( ':' );
				s.append( count[ i ] );
			}

		}
		return s.toString();
	}

	public int id() {
		return id;
	}

	/** The document uri. It may be null if it is not available. */
	public URI uri() {
		return uri;
	}

	public int size() {
		return numberOfTerms;
	}

	public int indexOf( int t ) {
		/*
		 * NOTE: Java 6 offers an overloaded Arrays.binarySearch with extremal specification.
		 */
		final int[] term = this.term;

		if ( sorted ) {
			int low = 0;
			int high = numberOfTerms - 1;

			while ( high >= low ) {
				final int mid = ( low + high ) / 2;
				if ( t < term[ mid ] ) high = mid - 1;
				else if ( t == term[ mid ] ) return mid;
				else low = mid + 1;
			}
		}
		else for ( int i = 0; i < numberOfTerms; i++ )
			if ( term[ i ] == t ) return i;

		return -1;
	}

	public int length() {
		return docLength;
	}

	@Override
	public int hashCode() {
		int h = 42;
		if ( count != null ) for ( int i = numberOfTerms; i-- != 0; ) h = ( h * 31 ) ^ ( term[ i ] * count[ i ] );
		else for ( int i = numberOfTerms; i-- != 0; ) h = ( h * 31 ) ^ term[ i ];
		return h ^ id ^ docLength;
	}

	/**
	 * Compares this document summary to another object.
	 * 
	 * <p>Note that this method will try to be smart if at least one of the two summaries is
	 * sorted. If you plan on calling this method often, please consider using sorted summaries.
	 * 
	 * @return true if <code>o</code> is a {@link DocumentSummary} with the same id, length, size
	 * and term-count information.
	 */
	@Override
	public boolean equals( final Object o ) {

		if ( !( o instanceof DocumentSummary ) ) return false;

		final DocumentSummary d = (DocumentSummary)o;
		if ( d.id() != id ) return false;
		if ( d.length() != docLength ) return false;

		if ( d.size() != numberOfTerms ) return false;

		if ( d.sorted() && sorted() ) {
			for ( int i = numberOfTerms; i-- != 0; )
				if ( term[ i ] != d.term( i ) || count( i ) != d.count( i ) ) return false;
			return true;
		}

		if ( d.sorted() ) {
			int p;
			for ( int i = numberOfTerms; i-- != 0; )
				if ( ( p = d.indexOf( term[ i ] ) ) == -1 || count( i ) != d.count( p ) ) return false;
			return true;
		}

		int p;
		for ( int i = numberOfTerms; i-- != 0; )
			if ( ( p = indexOf( d.term( i ) ) ) == -1 || count( p ) != d.count( i ) ) return false;
		return true;
	}
}
