package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
import it.unimi.dsi.Util;
import it.unimi.di.archive4j.ArchiveLoader.PropertyKeys;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Coding;
import it.unimi.di.archive4j.SequentialBitstreamArchive.CompressionFlags.Component;
import it.unimi.di.archive4j.tool.Preprocess;
import it.unimi.dsi.bits.Fast;
import it.unimi.dsi.bits.LongArrayBitVector;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.longs.LongArrayList;
import it.unimi.dsi.fastutil.longs.LongSet;
import it.unimi.dsi.io.OutputBitStream;
import it.unimi.dsi.lang.MutableString;
import it.unimi.dsi.sux4j.bits.SparseRank;
import it.unimi.dsi.sux4j.util.EliasFanoMonotoneLongBigList;
import it.unimi.dsi.util.Properties;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Map;

import org.apache.commons.configuration.ConfigurationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cern.colt.Sorting;
import cern.colt.function.IntComparator;

// RELEASE-STATUS: DIST


/**
 * A writer for {@link SequentialBitstreamArchive} or {@link RandomAccessBitstreamArchive} archives.
 * 
 * <p>Summaries are stored in a bitstream. Each summary is preceded by its id (for a
 * {@link SequentialBitstreamArchive}) or by the gap (reduced by one) with the previous id (for a
 * {@link RandomAccessBitstreamArchive}). Then, we write the size and the length of the document.
 * 
 * <p>Terms are renumbered by descending frequency rank, and sorted so to be able to store gaps.
 * Counts follow, stored in <em>reverse</em> order again using gaps: of course, now gap can be
 * negative (albeit, due to the correlation between global and local counts, we expect them to be
 * small) so we pass them through {@link Fast#int2nat(int)}. Codes for writing the above data can
 * be selected into by passing a
 * {@linkplain SequentialBitstreamArchive.CompressionFlags set of flags} to the constructors.
 * 
 * <p>It is possible to override permutation behaviour by explicitly passing a term-to-rank
 * permutation and its inverse.
 * 
 * <p>All in all, we generate the following files: 
 * 
 * <dl>
 * 
 * <dt><samp><var>basename</var>.archive</samp>
 * 
 * <dd>The archive bitstream.
 * 
 * <dt><samp><var>basename</var>.permutation</samp>
 * 
 * <dd>The rank-to-term permutation (translates data in the archive to term numbers).
 * 
 * <dt><samp><var>basename</var>.offsets</samp>
 * 
 * <dd>For random-access archives, a serialised {@link EliasFanoMonotoneLongBigList} storing
 * the start of each summary.
 * 
 * <dt><samp><var>basename</var>.missing</samp>
 * 
 * <dd>For random-access archives, a serialised {@link SparseRank} storing the ids of the documents
 * missing from the archive. 
 * 
 * </dl>
 * 
 * <p>A final overall property file <samp><var>basename</var>.properties</samp> records
 * a few {@link PropertyKeys properties}.

 * @see RandomAccessBitstreamArchive
 * @see SequentialBitstreamArchive
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 * 
 */
public class BitstreamArchiveWriter<T extends DocumentSummary> extends ArchiveWriter<T> {

	@SuppressWarnings("unused")
	private static Logger LOGGER = LoggerFactory.getLogger( BitstreamArchiveWriter.class );

	/** The extension for the file of stats. */
	public static final String STATS_EXTENSION = ".stats";

	public static int CURRENT_VERSION = 3;

	/** Output bit stream for documental data. */
	private final OutputBitStream output;

	/** List accumulating bit pointers on where documents begin in {@link #output}. */
	private final LongArrayList offsets;

	/**
	 * A bit vector where the <code>i</code>-th bit is set if the document with id <code>i</code>
	 * is the in archive.
	 */
	private final LongArrayBitVector docsBitVector;

	/** The {@link LongSet} representation of {@link #docsBitVector} */
	private final LongSet docs;

	/** Basename for the generated archive. */
	private final CharSequence basename;

	/** Whether we are writing a random-access archive. */
	private final boolean randomAccess;

	/** Maps term to its rank position, as they are written in the archive. */
	private final int[] term2Rank;

	/** The number of terms in the collections ({@link #term2Rank term2Rank.length}). */
	private final int numberOfTerms;
	
	/** The coding map of this writer. */
	private final Map<CompressionFlags.Component, CompressionFlags.Coding> codings;

	/** The coding for summary sizes. */
	private final CompressionFlags.Coding sizeCoding;
	
	/** The coding for document lengths. If <code>null</code>, document lengths are not stored. */
	private final CompressionFlags.Coding lengthCoding;

	/** The coding for term gaps. */
	private final CompressionFlags.Coding termCoding;

	/** The coding for count gaps. If <code>null</code>, gaps are not stored.*/
	private final CompressionFlags.Coding countCoding;

	/** The number of documents appended so far. */
	private int numberOfDocuments;

	/** The length of the documents in the archive archive so far. */
	private long cumulativeDocLength;

	/** The id of the last appended document. */
	private int lastDocId = -1;

	/** A temporary array used to sort term numbers by rank. */
	private int[] termPerm;
	
	/** The number of term entries appended so far. */
	private long numberOfPostings;

	/** The number of bits used to write term indices. */
	private long bitsForTerms;

	/** The number of bits used to write term counts. */
	private long bitsForCounts;

	/** The number of bits used to write document lengths. */
	private long bitsForLengths;

	/** The number of bits used to write summary sizes. */
	private long bitsForSizes;

	/** The number of bits used to write ids or id gaps. */
	private long bitsForIds;
	
	
	/**
	 * Build the frequency-rank-to-term and term-to-frequency-rank permutations, mapping each
	 * frequency rank to the respective term and viceversa.
	 * 
	 * @return a two-element array containing the term-to-rank and the rank-to-term permutations,
	 * respectively.
	 */
	public static int[][] makeSortingPermutations( final int[] frequency ) {
		int rank2Term[] = new int[ frequency.length ];
		for ( int i = frequency.length; i-- != 0; )
			rank2Term[ i ] = i;

		Sorting.quickSort( rank2Term, 0, rank2Term.length, new IntComparator() {
			// Sort in descending order using the documental frequency.
			public int compare( final int a, final int b ) {
				return frequency[ b ] - frequency[ a ];
			}
		} );

		final int term2Rank[] = new int[ frequency.length ];
		for ( int i = rank2Term.length; i-- != 0; )
			term2Rank[ rank2Term[ i ] ] = i;

		return new int[][] { term2Rank, rank2Term };
	}

	/** Returns an array containing the inverse permutation of the source one. */
	protected static int[] invertPermutation( final int[] source ) {
		int[] dest = new int[ source.length ];
		for ( int i = 0; i < dest.length; i++ )
			dest[ source[ i ] ] = i;

		return dest;
	}

	/**
	 * Creates a new bitstream archive writer with given basename and frequency array.
	 * 
	 * <p>You have the option to create an archive marked as <em>sorted</em>. Such an archive
	 * will be loaded as a {@link RandomAccessBitstreamArchive} and allow random access. You must,
	 * however, guarantee that calls to {@link ArchiveWriter#append(DocumentSummary)} happen with
	 * increasing document ids.
	 * 
	 * @param basename the basename of the archive.
	 * @param frequency the array of frequencies.
	 * @param sorted if true, the resulting archive will be sorted.
	 */
	public BitstreamArchiveWriter( final CharSequence basename, final int[] frequency, final boolean sorted, final Map<Component, Coding> codings ) throws IOException {
		this( basename, makeSortingPermutations( frequency ), sorted, codings );
	}

	/**
	 * Creates a new bitstream archive writer with given basename and sorting permutations.
	 * 
	 * <p>This constructor can be useful if you want to call
	 * {@link #makeSortingPermutations(int[])} on your own and then free the memory occupied by the
	 * frequency array.
	 * 
	 * <p>You have the option to create an archive marked as <em>sorted</em>. Such an archive
	 * will be loaded as a {@link RandomAccessBitstreamArchive} and allow random access. You must,
	 * however, guarantee that calls to {@link ArchiveWriter#append(DocumentSummary)} happen with
	 * increasing document ids.
	 * 
	 * @param basename the basename of the archive.
	 * @param permutations two arrays containing the sorting-by-rank and the inverse term
	 * permutations (i.e., the output of {@link #makeSortingPermutations(int[])}).
	 * @param sorted if true, the resulting archive will be sorted.
	 * @see #BitstreamArchiveWriter(CharSequence, int[], boolean, Map)
	 */
	public BitstreamArchiveWriter( final CharSequence basename, final int[][] permutations, final boolean sorted, final Map<Component, Coding> codings ) throws IOException {
		this( basename, permutations[ 0 ], permutations[ 1 ], sorted, codings );
	}



	/**
	 * Creates a new bitstream archive writer with given basename and permutations.
	 * 
	 * @param basename the basename of the archive.
	 * @param term2Rank the permutation from term number to frequency ranks.
	 * @param rank2Term the inverse of <code>term2Rank</code>.
	 * @param randomAccess if true, the resulting archive will be accessible randomly; documents
	 * must be provided in increasing order.
	 * @see #BitstreamArchiveWriter(CharSequence, int[], boolean, Map)
	 */
	protected BitstreamArchiveWriter( final CharSequence basename, final int[] term2Rank, final int[] rank2Term, final boolean randomAccess,
			final Map<CompressionFlags.Component, CompressionFlags.Coding> codings ) throws IOException {
		this.basename = basename;
		this.randomAccess = randomAccess;
		this.term2Rank = term2Rank;
		this.codings = codings;

		lengthCoding = codings.get( Component.DOCLENGTHS );
		sizeCoding = codings.get( Component.SIZES );
		termCoding = codings.get( Component.TERMS );
		countCoding = codings.get( Component.COUNTS );

		output = new OutputBitStream( new FileOutputStream( new File( basename + SequentialBitstreamArchive.ARCHIVE_EXTENSION ) ) );
		offsets = new LongArrayList();
		offsets.add( 0 );
		docsBitVector = LongArrayBitVector.getInstance();
		docs = docsBitVector.asLongSet();
		numberOfTerms = term2Rank.length;
		termPerm = IntArrays.EMPTY_ARRAY;

		BinIO.storeInts( rank2Term, basename + SequentialBitstreamArchive.PERM_EXTENSION );
	}


	public BitstreamArchiveWriter( final String basename, SequentialBitstreamArchive prototype, boolean randomAccess ) throws IOException {
		this( basename, prototype, randomAccess, prototype.getCodings() );

	}

	public BitstreamArchiveWriter( final String basename, SequentialBitstreamArchive prototype, boolean randomAccess, final Map<Component, Coding> codings ) throws IOException {
		this( basename, new int[][] { invertPermutation( prototype.getPermutation() ), prototype.getPermutation() }, randomAccess, codings );

	}


	private final int writeInt( final OutputBitStream stream, final int x, final Coding coding ) throws IOException {
		switch ( coding ) {
		case UNARY:
			return stream.writeUnary( x );
		case GAMMA:
			return stream.writeGamma( x );
		case DELTA:
			return stream.writeDelta( x );
		case SHIFTED_GAMMA:
			return stream.writeShiftedGamma( x );
		case NIBBLE:
			return stream.writeNibble( x );
		case ZETA_2:
			return stream.writeZeta( x, 2 );
		case ZETA_3:
			return stream.writeZeta( x, 3 );
		case ZETA_4:
			return stream.writeZeta( x, 4 );
		case ZETA_5:
			return stream.writeZeta( x, 5 );
		default:
			throw new IllegalStateException( "The encoding " + coding + "is not supported." );
		}
	}

	@Override
	public void append( final T summary ) throws IOException {

		if ( summary.id() < lastDocId && randomAccess ) throw new IllegalStateException( "This writer produces sorted archives but input documents are unsorted.( " + summary.id() + " < " + lastDocId );

		bitsForIds += output.writeDelta( randomAccess ? summary.id() - lastDocId - 1 : summary.id() );
		lastDocId = summary.id();
		docs.add( lastDocId );

		final int t = summary.size();
		numberOfPostings += t;
		final int[] term2Rank = this.term2Rank;

		final int[] termPerm;

		// Produce the permutation termPerm of {0,..,t-1} so that termPerm[0] is term2Rank[term[0]]
		if ( term2Rank != null ) {
			termPerm = this.termPerm = IntArrays.grow( this.termPerm, t );
			for ( int i = t; i-- != 0; )
				termPerm[ i ] = i;

			Sorting.quickSort( termPerm, 0, t, new IntComparator() {
				public int compare( final int a, final int b ) {
					return term2Rank[ summary.term( a ) ] - term2Rank[ summary.term( b ) ];
				}
			} );
		}
		else termPerm = null;

		final int len = summary.length();
		cumulativeDocLength += len;
		if ( lengthCoding != null ) bitsForLengths += writeInt( output, len, lengthCoding );

		bitsForSizes += writeInt( output, t, sizeCoding );

		for ( int i = 0, term, prevTerm = -1; i < t; i++ ) {
			term = term2Rank == null ? summary.term( i ) : term2Rank[ summary.term( termPerm[ i ] ) ];
			bitsForTerms += writeInt( output, term - prevTerm - 1, termCoding );
			prevTerm = term;
		}

		if ( countCoding != null && t != 0 ) {
			int oldCount = 0, count = 0, pos = term2Rank == null ? t - 1 : termPerm[ t - 1 ];
			bitsForCounts += writeInt( output, ( oldCount = summary.count( pos ) ) - 1, countCoding );

			for ( int i = 1; i < t; i++ ) {
				pos = term2Rank == null ? t - i - 1 : termPerm[ t - i - 1 ];
				count = summary.count( pos );
				bitsForCounts += writeInt( output, Fast.int2nat( count - oldCount ), countCoding );
				oldCount = count;
			}
		}
		
		numberOfDocuments++;

		if ( randomAccess ) offsets.add( output.writtenBits() );
	}

	public void close() throws IOException {
		if ( output.writtenBits() != bitsForCounts + bitsForIds + bitsForLengths + bitsForSizes + bitsForTerms ) throw new IllegalStateException( output.writtenBits() + " != " + ( bitsForCounts + bitsForIds + bitsForLengths + bitsForSizes + bitsForTerms ) );
		output.close();

		if ( randomAccess ) {
			BinIO.storeObject( new EliasFanoMonotoneLongBigList( numberOfDocuments + 1, output.writtenBits() + 1, offsets.iterator() ), basename + RandomAccessBitstreamArchive.OFFSETS_EXTENSION );
			assert docsBitVector.length() == lastDocId + 1;
			docsBitVector.flip();
			BinIO.storeObject( new SparseRank( docsBitVector ), basename + RandomAccessBitstreamArchive.MISSING_EXTENSION );
		}

		try {
			String propertyFilename = basename + Preprocess.PROPERTIES_EXTENSION;

			Properties properties = new File( propertyFilename ).exists() ? new Properties( propertyFilename ) : new Properties();
			properties.setProperty( PropertyKeys.DOCUMENTS, numberOfDocuments );
			properties.setProperty( PropertyKeys.TERMS, numberOfTerms );
			properties.setProperty( PropertyKeys.WORDS, cumulativeDocLength );
			properties.setProperty( PropertyKeys.CLASS, ( randomAccess ? RandomAccessBitstreamArchive.class : SequentialBitstreamArchive.class ).getName() );

			for ( Map.Entry<Component, Coding> e : codings.entrySet() )
				if ( e.getValue() != null ) properties.addProperty( SequentialBitstreamArchive.PropertyKeys.CODING, new MutableString().append( e.getKey() ).append( ':' ).append( e.getValue() ) );

			properties.setProperty( SequentialBitstreamArchive.PropertyKeys.VERSION, CURRENT_VERSION );
			properties.save( propertyFilename );
		}
		catch ( ConfigurationException e ) {
			throw new RuntimeException( e );
		}
		
		final PrintWriter pw = new PrintWriter( new FileWriter ( basename + STATS_EXTENSION ) );
		pw.println( "Number of documents: " + Util.format( numberOfDocuments ) );
		pw.println( "Number of terms: " + Util.format( numberOfTerms ) );
		pw.println( "Number of postings: " + Util.format( numberOfPostings ) );
		pw.println( "Number of words (term occurrences): " + Util.format( cumulativeDocLength ) );
		pw.println( "Size: " + Util.format( output.writtenBits() ) + " bits (" + Util.format( output.writtenBits() / (double)numberOfDocuments ) + " bits per summary)" );
		pw.println( "Bits for ids: " + Util.format( bitsForIds ) + " (" + Util.format( bitsForIds / (double)numberOfDocuments ) + " bits per id)" );
		pw.println( "Bits for lenghts: " + Util.format( bitsForLengths ) + " (" + Util.format( bitsForLengths / (double)numberOfDocuments ) + " bits per length)" );
		pw.println( "Bits for sizes: " + Util.format( bitsForSizes ) + " (" + Util.format( bitsForSizes / (double)numberOfDocuments ) + " bits per size)" );
		pw.println( "Bits for terms: " + Util.format( bitsForTerms ) + " (" + Util.format( bitsForTerms / (double)numberOfPostings ) + " bits per entry)" );
		pw.println( "Bits for counts: " + Util.format( bitsForCounts ) + " (" + Util.format( bitsForCounts / (double)numberOfPostings ) + " bits per count)" );
		pw.close();
	}


}
