package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.io.InputBitStream;
import it.unimi.dsi.sux4j.bits.SimpleSelectZero;
import it.unimi.dsi.sux4j.bits.SparseRank;
import it.unimi.dsi.sux4j.util.EliasFanoMonotoneLongBigList;
import it.unimi.dsi.util.Properties;
import it.unimi.dsi.util.StringMap;

import java.io.Closeable;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;




/**
 * An {@link Archive} implementation providing random access.
 * 
 * <p>Each instance uses a single {@link InputBitStream}, so you must not mix sequential and
 * random accesses.
 * 
 * @see BitstreamArchiveWriter
 * @see SequentialBitstreamArchive
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */

public class RandomAccessBitstreamArchive extends SequentialBitstreamArchive implements Closeable {
	@SuppressWarnings("unused")
	private static final Logger LOGGER = LoggerFactory.getLogger( RandomAccessBitstreamArchive.class );

	/** The extension for offset file */
	public static final String OFFSETS_EXTENSION = ".offsets";

	/** The extension for the missing-document file. */
	public static final String MISSING_EXTENSION = ".missing";

	/** A rank structure whose underlying bit vector marks the ids of the documents missing from the archive. */
	protected SparseRank id2Index;

	/**
	 * A selection structure on the same array as {@link #id2Index}. It is initialised lazily upon
	 * the first access to {@link #getDocumentByIndex(int)}
	 */
	protected transient SimpleSelectZero index2id;

	/** A selection structure storing the bit offsets of each document. */
	protected EliasFanoMonotoneLongBigList offsets;

	/** Creates a new random-access bitstream archive.
	 * 
	 * @param basename the basename of the archive.
	 * @param offsets the list of offsets, pointing at the start of each summary.
	 * @param id2Index a ranking structure returning the index of a document given its id. 
	 * @param rank2Term the permutation from rank to terms.
	 * @param properties the properties of the archive.
	 * @param uriList an optional list of URIs that will be used to associate a URI to each summary, or <code>null</code>.
	 * @param frequency the term frequencies.
	 */
	protected RandomAccessBitstreamArchive( final CharSequence basename, EliasFanoMonotoneLongBigList offsets, SparseRank id2Index, int[] rank2Term, Properties properties, List<? extends CharSequence> uriList, final int[] frequency ) throws IOException {
		super( basename, rank2Term, properties, uriList, frequency );
		this.offsets = offsets;
		this.id2Index = id2Index;
	}

	@Override
	public ArrayDocumentSummary getDocumentById( final int id ) throws IOException {
		ensureOpen();
		if ( id < 0 ) throw new IndexOutOfBoundsException( Integer.toString( id ) );
		final int rankZero = (int)id2Index.rankZero( id );
		if ( id2Index.rankZero( id + 1 ) == rankZero ) return null;
		if ( rankZero < 0 || rankZero >= numberOfDocuments ) throw new IndexOutOfBoundsException( Integer.toString( rankZero ) );
		data.position( offsets.getLong( rankZero ) );
		data.readDelta(); // Skip offset delta
		return readCurrentDocument( id );
	}

	@Override
	public ArrayDocumentSummary getDocumentByIndex( final int index ) throws IOException {
		ensureOpen();
		if ( index < 0 || index >= numberOfDocuments ) throw new IndexOutOfBoundsException( Integer.toString( index ) );
		if ( index2id == null ) index2id = new SimpleSelectZero( id2Index.bitVector() );
		data.position( offsets.getLong( index ) );
		data.readDelta(); // Skip id delta
		return readCurrentDocument( (int)index2id.selectZero( index ) );
	}

	@Override
	public boolean hasRandomAccess() {
		return true;
	}

	protected RandomAccessBitstreamArchive( final RandomAccessBitstreamArchive prototype ) throws IOException {
		super( prototype );
		this.index2id = prototype.index2id;
		this.id2Index = prototype.id2Index;
		this.offsets = prototype.offsets;
	}


	/** This methods returns an efficient iterator. */
	@Override
	public Iterator<ArrayDocumentSummary> iterator() {

		ensureOpen();

		try {
			data.position( 0 );
		}
		catch ( IOException e ) {
			throw new RuntimeException( e );
		}

		return new AbstractObjectIterator<ArrayDocumentSummary>() {
			private int currIndex = 0;

			private int lastId = -1;

			public boolean hasNext() {
				return currIndex < numberOfDocuments;
			}

			public ArrayDocumentSummary next() {
				ensureOpen();
				if ( !hasNext() ) throw new NoSuchElementException();

				try {
					currIndex++;
					lastId += data.readDelta() + 1;
					return readCurrentDocument( lastId );
				}
				catch ( IOException e ) {
					throw new RuntimeException( e );
				}

			}
		};
	}


	@Override
	public RandomAccessBitstreamArchive copy() {
		try {
			return new RandomAccessBitstreamArchive( this );
		}
		catch ( IOException e ) {
			throw new RuntimeException( e );
		}
	}

	/** Returns a {@link RandomAccessBitstreamArchive} obtained by loading with given basename and optional URI list.
	 * 
	 * @param basename the archive basename.
	 * @param properties the archive properties.
	 * @param uriFilename the filename of a URI list, or <code>null</code>; the file must contained either
	 * a {@link StringMap} supporting {@link StringMap#list()}, or a {@link List} of {@link CharSequence}s.
	 * @return the {@link RandomAccessBitstreamArchive} with given basename and URI list.
	 */

	@SuppressWarnings("unchecked")
	public static RandomAccessBitstreamArchive getInstance( final CharSequence basename, final Properties properties, final CharSequence uriFilename ) throws IOException, ClassNotFoundException {
		final Object uriData = uriFilename == null ? null : BinIO.loadObject( uriFilename );
		return new RandomAccessBitstreamArchive( basename, (EliasFanoMonotoneLongBigList)BinIO.loadObject( basename + OFFSETS_EXTENSION ),
				(SparseRank)BinIO.loadObject( basename + MISSING_EXTENSION ), BinIO.loadInts( basename + PERM_EXTENSION ), properties,
				uriData == null ? null : uriData instanceof StringMap ? ((StringMap<? extends CharSequence>)uriData).list() : (List<? extends CharSequence>)uriData, loadFrequencies( basename, properties.getInt( ArchiveLoader.PropertyKeys.TERMS ) ) );
	}
}
