package it.unimi.di.archive4j;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.bits.LongArrayBitVector;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.longs.LongSortedSet;
import it.unimi.dsi.io.InputBitStream;
import it.unimi.dsi.io.OutputBitStream;
import it.unimi.dsi.sux4j.bits.SparseRank;
import it.unimi.dsi.sux4j.bits.SparseSelect;
import it.unimi.dsi.util.Properties;
import it.unimi.dsi.util.StringMap;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Iterator;

import org.apache.commons.configuration.ConfigurationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;

/**
 * A generic subarchive, thought as a filter over existing IDs over the parent one. <p>It uses a
 * {@link SparseSelect} object to maintain a list of the IDs that are kept from the parent archive,
 * which must be random accessible.
 * 
 * <p>During construction, it may also recover the true frequency of each term in the included
 * documents.
 * 
 * @author Alessio Orlandi
 */

public class Subarchive<T extends DocumentSummary> implements Archive<T> {
	private static final Logger LOGGER = LoggerFactory.getLogger( Subarchive.class );

	/** The select filter data file. */
	public static final String FILTER_EXTENSION = ".filter";

	/** The (optional) frequencies data file */
	public static final String FREQUENCIES_EXTENSION = ".frequencies";

	/** @see ArchiveLoader.PropertyKeys */
	public enum PropertyKeys {
		/** The basename for the parent archive. */
		PARENT_BASENAME
	}


	/**
	 * The specific writer, including useful extra append calls to build archives without opening
	 * the parent one.
	 */

	public static class SubarchiveWriter<T extends DocumentSummary> extends ArchiveWriter<T> {

		private Archive<T> parentArchive;

		private LongArrayBitVector vector;

		private LongSortedSet documents;

		private CharSequence parentBasename, outputBasename;

		private int[] frequency;

		private int cumulativeWordLength;

		@SuppressWarnings("unchecked")
		public SubarchiveWriter( CharSequence outputBasename, CharSequence parentBasename, boolean storeFrequencies ) throws IllegalArgumentException {
			this.vector = LongArrayBitVector.getInstance();
			this.documents = vector.asLongSet();
			this.parentBasename = parentBasename;
			this.outputBasename = outputBasename;
			try {
				this.parentArchive = (Archive<T>)ArchiveLoader.getInstance( parentBasename );
			}
			catch ( Exception e ) {
				throw new IllegalArgumentException( "Cannot open the parent archive" + e );
			}

			if ( !this.parentArchive.hasRandomAccess() ) throw new IllegalArgumentException( "The parent archive doesn't support random access as required." );

			this.frequency = storeFrequencies ? new int[ parentArchive.numberOfTerms() ] : null;
			this.cumulativeWordLength = 0;
		}

		@Override
		public void append( T document ) throws IOException {
			documents.add( document.id() );

			cumulativeWordLength += document.length();
			if ( frequency != null ) for ( int i = document.size(); --i >= 0; )
				frequency[ document.term( i ) ]++;
		}

		/** Appends a new document ID, without checking its existance, to the filter. */
		public void appendID( int id ) throws IOException {
			append( parentArchive.getDocumentById( id ) );
		}

		public void close() throws IOException {
			BinIO.storeObject( new SparseSelect( vector ), outputBasename + Subarchive.FILTER_EXTENSION );

			if ( frequency != null ) {
				final OutputBitStream freqobs = new OutputBitStream( outputBasename + FREQUENCIES_EXTENSION );
				for ( int f : frequency )
					freqobs.writeGamma( f );
				freqobs.close();
			}

			final Properties props = new Properties();
			props.setProperty( PropertyKeys.PARENT_BASENAME, parentBasename );
			props.setProperty( ArchiveLoader.PropertyKeys.CLASS, Subarchive.class.getName() );
			props.setProperty( ArchiveLoader.PropertyKeys.TERMS, parentArchive.numberOfTerms() );
			props.setProperty( ArchiveLoader.PropertyKeys.WORDS, cumulativeWordLength );
			props.setProperty( ArchiveLoader.PropertyKeys.DOCUMENTS, documents.size() );

			try {
				props.save( outputBasename + ArchiveLoader.PROPERTIES_EXTENSION );
			}
			catch ( ConfigurationException e ) {
				throw new RuntimeException( e );
			}

			parentArchive.close();
		}
	}


	/** The object representing the parent archive. */
	protected final Archive<T> parentArchive;

	/** Number of documents of the superarchive exposed by this archive. */
	protected final int numberOfDocuments;

	/**
	 * A dictionary containing document IDs that are contained in this subarchive.
	 */
	protected final SparseSelect filter;

	/** The ranking data structure exposed by {@link #filter}. */
	protected final SparseRank filterRank;

	/** The frequency for all those documents in this archive, or <code>null</code>. */
	protected final int[] frequency;

	/** The cumulative word document length for all those documents in this archive. */
	protected final int cumulativeWordLength;

	/** Set to true when {@link #close()} is called. */
	protected boolean closed = false;

	public Subarchive( Archive<T> parentArchive, SparseSelect filter, int[] frequency, int cumulativeWordLength, int numberOfDocuments ) {
		this.filter = filter;
		this.parentArchive = parentArchive;
		this.frequency = frequency;
		this.cumulativeWordLength = cumulativeWordLength;
		this.filterRank = filter.getRank();
		this.numberOfDocuments = numberOfDocuments;
	}

	protected Subarchive( Archive<T> parentArchive, SparseSelect filter, int[] frequency, Properties props ) {
		this( parentArchive, filter, frequency, props.getInt( ArchiveLoader.PropertyKeys.WORDS ), props.getInt( ArchiveLoader.PropertyKeys.DOCUMENTS ) );
	}

	public T getDocumentById( int id ) throws IOException {
		if ( closed ) throw new IllegalStateException( "The archive has been closed." );

		long pos = filterRank.rank( id );

		if ( pos == filterRank.rank( id + 1 ) ) {
			return null;
		}

		return parentArchive.getDocumentById( id );
	}

	public T getDocumentByIndex( int idx ) throws IOException {
		if ( closed ) throw new IllegalStateException( "The archive has been closed." );

		final int id = (int)filter.select( idx );
		return ( id >= 0 ? parentArchive.getDocumentById( id ) : null );
	}

	public Iterator<T> iterator() {
		if ( closed ) throw new IllegalStateException( "The archive has been closed." );

		return new Iterator<T>() {

			int t = 0;

			public boolean hasNext() {
				return t < numberOfDocuments;
			}

			public T next() {
				try {
					return getDocumentByIndex( t++ );
				}
				catch ( IOException e ) {
					throw new RuntimeException( e );
				}
			}

			public void remove() {
				throw new UnsupportedOperationException();
			}
		};
	}

	public int numberOfDocuments() {
		return numberOfDocuments;
	}

	public int numberOfTerms() {
		return parentArchive.numberOfTerms();
	}

	public void close() throws IOException {
		if ( !closed ) parentArchive.close();
		closed = true;
	}

	public int frequency( int term ) {
		if ( frequency == null ) throw new UnsupportedOperationException();
		return frequency[ term ];
	}

	public long numberOfWords() {
		return cumulativeWordLength;
	}

	/** Returns the parent archive. */
	public Archive<T> parent() {
		return parentArchive;
	}

	public boolean hasRandomAccess() {
		return true;
	}

	/** Returns a flyweight copy of this object if the parent archive supports it. */
	public Subarchive<T> copy() {

		if ( closed ) throw new IllegalStateException();

		return new Subarchive<T>( parentArchive.copy(), filter, frequency, cumulativeWordLength, numberOfDocuments );
	}

	@Override
	protected void finalize() throws Throwable {
		try {
			if ( !closed ) {
				LOGGER.warn( "Subarchive should have been closed" );
				close();
			}
		}
		finally {
			super.finalize();
		}
	}


	@SuppressWarnings("unchecked")
	public static <T extends DocumentSummary> Subarchive<T> getInstance( CharSequence basename, Properties properties, CharSequence urilist ) throws IOException, ConfigurationException,
			ClassNotFoundException {
		final CharSequence parent = properties.getString( PropertyKeys.PARENT_BASENAME );
		final Archive<T> parentArchive = (Archive<T>)ArchiveLoader.getInstance( parent, urilist );

		final SparseSelect filter = (SparseSelect)BinIO.loadObject( basename + FILTER_EXTENSION );
		int[] frequency = null;

		if ( new File( basename + FREQUENCIES_EXTENSION ).exists() ) {
			final InputBitStream ibs = new InputBitStream( basename + FREQUENCIES_EXTENSION );
			frequency = new int[ properties.getInt( ArchiveLoader.PropertyKeys.TERMS ) ];
			ibs.readGammas( frequency, frequency.length );
		}

		return new Subarchive<T>( parentArchive, filter, frequency, properties );
	}

	/**
	 * Commodity method returning an object of this class containing documents from the given list
	 * that appear in the parent archive. Mapping from strings to document ID is performed by the
	 * given map.
	 */
	public static <T extends DocumentSummary> Subarchive<T> getInstanceWithList( Archive<T> parent, StringMap<CharSequence> map, Iterable<? extends CharSequence> list ) throws IOException {
		final LongArrayBitVector vector = LongArrayBitVector.getInstance();
		final LongSortedSet documents = vector.asLongSet();

		int id, count = 0, wlength = 0;
		for ( CharSequence uri : list )
			if ( ( id = (int)map.getLong( uri ) ) >= 0 ) {
				final T docObj = parent.getDocumentById( id );
				if ( docObj != null ) {
					documents.add( id );
					count++;
					wlength += docObj.length();
				}
			}

		return new Subarchive<T>( parent, new SparseSelect( vector ), null, wlength, count );
	}

	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {
		final SimpleJSAP jsap = new SimpleJSAP( Subarchive.class.getName(), "Creates a subarchive by including only those urls fed via stdin that appear in the given map", new Parameter[] {
				new Switch( "frequencies", 'f', "frequencies", "Build term frequencies (requires document analysis)" ),
				new FlaggedOption( "urlMap", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, 'u', "urlMap", "A map from URIs to document identifiers that will be used to renumber documents." ),
				new FlaggedOption( "input", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, 'i', "input", "The input basename archive." ),
				new UnflaggedOption( "output", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The archive output basename." ), } );

		final JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;


		final BufferedReader br = new BufferedReader( new InputStreamReader( System.in ) );
		String s;

		final StringMap<CharSequence> map = (StringMap<CharSequence>)BinIO.loadObject( jsapResult.getString("urlMap") );

		final SubarchiveWriter writer = new SubarchiveWriter( jsapResult.getString( "output" ), jsapResult.getString( "input" ), jsapResult.getBoolean( "frequencies" ) );
		final Archive<? extends DocumentSummary> a = ArchiveLoader.getInstance(jsapResult.getString("input"));
		
		while ( ( s = br.readLine() ) != null ) {
			try {
				long id;
				if ( ( id = map.getLong( s ) ) != -1  && a.getDocumentById( (int)id ) != null ) writer.appendID( (int)id );
			}
			catch ( IndexOutOfBoundsException ignored ) {}
		}

		writer.close();
		a.close();
	}


}
