package it.unimi.di.archive4j.tool;

/*
 * Copyright (C) 2008-2013 Alessio Orlandi and Sebastiano Vigna
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.di.big.mg4j.document.Document;
import it.unimi.di.big.mg4j.document.DocumentCollection;
import it.unimi.di.big.mg4j.document.DocumentIterator;
import it.unimi.di.big.mg4j.document.DocumentSequence;
import it.unimi.di.big.mg4j.document.IdentityDocumentFactory;
import it.unimi.di.big.mg4j.index.NullTermProcessor;
import it.unimi.di.big.mg4j.index.TermProcessor;
import it.unimi.di.big.mg4j.tool.Scan;
import it.unimi.di.big.mg4j.util.MG4JClassParser;
import it.unimi.dsi.Util;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.io.FastBufferedOutputStream;
import it.unimi.dsi.fastutil.longs.AbstractLongIterator;
import it.unimi.dsi.fastutil.objects.Object2ReferenceMap;
import it.unimi.dsi.fastutil.objects.Object2ReferenceOpenHashMap;
import it.unimi.dsi.fastutil.objects.ObjectIterator;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;
import it.unimi.dsi.io.OutputBitStream;
import it.unimi.dsi.io.WordReader;
import it.unimi.dsi.lang.MutableString;
import it.unimi.dsi.lang.ObjectParser;
import it.unimi.dsi.logging.ProgressLogger;
import it.unimi.dsi.util.Properties;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Reader;

import org.apache.commons.configuration.ConfigurationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cern.colt.GenericSorting;
import cern.colt.Swapper;
import cern.colt.function.IntComparator;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.UnflaggedOption;

/** Analyses a {@linkplain DocumentSequence} and generates data used by {@link MergePreprocessedData} and {@link Scan}.
 * Each document is analysed, and its terms are passed through a {@linkplain TermProcessor}.
 *  
 * <p>We compute a number of <em>batches</em>, formed 
 * by the following files:
 * 
 * <dl>
 * 
 * <dt><samp><var>basename</var>@<var>n</var>.terms</samp>
 * 
 * <dd>The lexicographically sorted list of terms in the batch.
 * 
 * <dt><samp><var>basename</var>@<var>n</var>.lengths</samp>
 * 
 * <dd>The lengths of the documents of the batch.
 * 
 * <dt><samp><var>basename</var>@<var>n</var>.frequencies</samp>
 * 
 * <dd>The frequency of each term in the batch.
 * 
 * <dt><samp><var>basename</var>@<var>n</var>.counts</samp>
 * 
 * <dd>The global count of each term in the batch.
 *
 * </dl>
 * 
 * <p>A final overall property file <samp><var>basename</var>.properties</samp> records
 * a few {@link PropertyKeys properties}.
 * 
 * @author Alessio Orlandi
 * @author Sebastiano Vigna
 */

public class Preprocess {
	private static final Logger LOGGER = LoggerFactory.getLogger( Preprocess.class );

	/** The minimum number of words that must be scanned between two consecutive memory checks (to avoid excessive calls to {@link Runtime#freeMemory()}). */
	private static final long MIN_WORDS_BETWEEN_MEMORY_CHECKS = 100000;

	/** The extension for the file of term frequencies. */
	public static final String FREQUENCIES_EXTENSION = ".frequencies";

	/** The extension for the file of document lengths. */
	public static final String LENGTHS_EXTENSION = ".lengths";

	/** The extension for the gzip'd term file. */
	public static final String TERMS_EXTENSION = ".terms";

	/** The extension for file of global counts. */
	public static final String COUNTS_EXTENSION = ".counts";

	/** The extension for the property file. */
	public static final String PROPERTIES_EXTENSION = ".properties";

	public static enum PropertyKeys {
		/** The term processor name. */
		TERMPROCESSOR,
		/** The indexed field. */
		FIELD,
		/** The number of batches that have been generated. */
		BATCHES,
		/** Number of documents in the archive. */
		DOCUMENTS
	};

	/**
	 * Statistical data associated to each term.
	 */
	final private static class TermStatistics {
		/** Number of documents in which it appears at least one time. */
		public int frequency;

		/** Number of occurrences for the term in the whole collection. */
		public long globalCount;

		public TermStatistics( int globalCount ) {
			this.frequency = 1;
			this.globalCount = globalCount;
		}

	}

	private final static int INITIAL_SIZE = 10000, INITIAL_SET_SIZE = 256;

	/**
	 * Returns the basename of a batch.
	 * 
	 * @param basename the basename of the archive.
	 * @param batch the batch number.
	 * @return the basename for the given batch.
	 */
	public static String batchName( final String basename, final int batch ) {
		return basename + "@" + batch;
	}

	/** Emits the current batch sorting and processing terms. */
	protected static void emit( final String batchName, final TermProcessor processor, final Object2ReferenceOpenHashMap<MutableString, TermStatistics> globalTerms ) throws IOException {

		// Sort term-processed terms.
		final MutableString[] sortedTerms = new MutableString[ globalTerms.size() ];
		final TermStatistics[] stats = new TermStatistics[ globalTerms.size() ];

		int cnt = 0;
		for ( ObjectIterator<Object2ReferenceMap.Entry<MutableString, TermStatistics>> i = globalTerms.object2ReferenceEntrySet().fastIterator(); i.hasNext(); ) {
			final Object2ReferenceMap.Entry<MutableString, TermStatistics> e = i.next();

			sortedTerms[ cnt ] = e.getKey();
			stats[ cnt ] = e.getValue();

			if ( processor.processTerm( sortedTerms[ cnt ] ) ) cnt++;
		}

		globalTerms.clear();

		GenericSorting.quickSort( 0, cnt, new IntComparator() {
			public int compare( int x, int y ) {
				return sortedTerms[ x ].compareTo( sortedTerms[ y ] );
			}
		}, new Swapper() {
			public void swap( int x, int y ) {
				final MutableString stmp = sortedTerms[ x ];
				sortedTerms[ x ] = sortedTerms[ y ];
				sortedTerms[ y ] = stmp;

				final TermStatistics tmp = stats[ x ];
				stats[ x ] = stats[ y ];
				stats[ y ] = tmp;
			}
		} );

		// Emit the batch so that MergeTermMaps can read them.
		final OutputBitStream freqdata = new OutputBitStream( new File( batchName + FREQUENCIES_EXTENSION ) );
		for ( int i = 0; i < cnt; i++ )
			freqdata.writeGamma( stats[ i ].frequency );
		freqdata.close();

		BinIO.storeLongs( new AbstractLongIterator() {
			final int len = stats.length;

			int ii = 0;

			@Override
			public long nextLong() {
				return stats[ ii++ ].globalCount;
			}

			public boolean hasNext() {
				return ii < len;
			}

		}, batchName + COUNTS_EXTENSION );

		final PrintWriter pw = new PrintWriter( new OutputStreamWriter( new FastBufferedOutputStream( new FileOutputStream( batchName + TERMS_EXTENSION ) ), "UTF-8" ) );
		for ( int i = 0; i < cnt; i++ )
			pw.println( sortedTerms[ i ] );
		pw.close();
	}


	/**
	 * Preprocesses the given document sequence.
	 * 
	 * @param basename the output basename.
	 * @param sequence the document sequence to collect terms from.
	 * @param processor the term processor to apply.
	 * @param indexedField the field to be indexed.
	 */
	public static void run( final String basename, final DocumentSequence sequence, final TermProcessor processor, final String indexedField ) throws IOException, ConfigurationException {

		final ProgressLogger pl = new ProgressLogger( LOGGER );

		pl.logger.info( "Processing terms with a " + processor.getClass().getName() );

		final DocumentIterator iterator = sequence.iterator();
		final Object2ReferenceOpenHashMap<MutableString, TermStatistics> terms = new Object2ReferenceOpenHashMap<MutableString, TermStatistics>( INITIAL_SIZE, Hash.FAST_LOAD_FACTOR );;
		final ObjectOpenHashSet<MutableString> termsForDocument = new ObjectOpenHashSet<MutableString>( INITIAL_SET_SIZE, Hash.VERY_FAST_LOAD_FACTOR );
		final MutableString word = new MutableString(), nonWord = new MutableString();
		// Contains the word length of all documents in delta coding.
		final OutputBitStream lengths = new OutputBitStream( basename + LENGTHS_EXTENSION );

		int batch = 0; // The index of the current batch.

		final int field = sequence.factory().fieldIndex( indexedField );
		if ( field == -1 ) throw new IllegalArgumentException( "Unknown field " + indexedField );

		boolean stopCompaction = false;

		if ( sequence instanceof DocumentCollection ) pl.expectedUpdates = ( (DocumentCollection)sequence ).size();
		pl.itemsName = "documents";
		pl.displayFreeMemory = true;
		pl.start( "Preprocessing field " + indexedField + "..." );

		Document d;
		int numberOfDocuments = 0;
		long wordsSinceLastMemoryCheck = 0;
		
		while ( ( d = iterator.nextDocument() ) != null ) {
			numberOfDocuments++;
			int len = 0;
			termsForDocument.clear();
			WordReader wordReader = d.wordReader( field );
			wordReader.setReader( (Reader)d.content( field ) );

			while ( wordReader.next( word, nonWord ) ) {
				if ( word.length() == 0 ) continue;
				len++;

				final TermStatistics stats = terms.get( word );
				if ( !termsForDocument.contains( word ) ) {
					final MutableString copy = word.copy();
					termsForDocument.add( copy );
					if ( stats == null ) terms.put( copy, new TermStatistics( 1 ) );
					else {
						stats.frequency++;
						stats.globalCount++;
					}
				}
				else stats.globalCount++;
			}

			lengths.writeDelta( len );
			wordsSinceLastMemoryCheck += len;
			d.close();

			if ( pl != null ) pl.update();

			if ( wordsSinceLastMemoryCheck > MIN_WORDS_BETWEEN_MEMORY_CHECKS ) {
				wordsSinceLastMemoryCheck = 0;
				long percAvailableMemory = Util.percAvailableMemory();
				if ( percAvailableMemory < 25 && !stopCompaction ) {
					if ( pl != null ) pl.logger.info( "Trying compaction... (" + percAvailableMemory + "% available)" );
					Util.compactMemory();
					percAvailableMemory = Util.percAvailableMemory();
					if ( percAvailableMemory < 35 ) stopCompaction = true;
					if ( pl != null ) pl.logger.info( "Compaction completed (" + percAvailableMemory + "% available" + ( stopCompaction ? ")" : ", will try again)" ) );
				}

				if ( percAvailableMemory < 25 ) {
					if ( pl != null ) pl.logger.warn( "Available memory below 25% : writing  batch # " + batch + " of " + terms.size() + " terms" );
					emit( batchName( basename, batch++ ), processor, terms );
					stopCompaction = false;
				}
			}
			
		}

		if ( terms.size() > 0 ) emit( batchName( basename, batch++ ), processor, terms );

		iterator.close();
		lengths.close();
		sequence.close();
		if ( pl != null ) pl.done();

		Properties properties = new Properties();
		properties.setProperty( PropertyKeys.DOCUMENTS, numberOfDocuments );
		properties.setProperty( PropertyKeys.FIELD, indexedField );
		properties.setProperty( PropertyKeys.TERMPROCESSOR, ObjectParser.toSpec( processor ) );
		properties.setProperty( PropertyKeys.BATCHES, batch );
		properties.save( basename + PROPERTIES_EXTENSION );
	}


	@SuppressWarnings("unchecked")
	public static void main( String[] args ) throws Exception {
		final TermProcessor termProcessor;

		final SimpleJSAP jsap = new SimpleJSAP( Preprocess.class.getName(), "Preprocesses a document sequence, generating data and statistics.", new Parameter[] {
						new FlaggedOption( "sequence", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'S', "sequence", "A serialised document sequence that will be used instead of stdin." ),
						new FlaggedOption( "objectSequence", new ObjectParser( DocumentSequence.class, MG4JClassParser.PACKAGE ), JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'o', "object-sequence",
								"An object specification describing a document sequence that will be used instead of stdin." ),
						new FlaggedOption( "delimiter", JSAP.INTEGER_PARSER, Integer.toString( Scan.DEFAULT_DELIMITER ), JSAP.NOT_REQUIRED, 'd', "delimiter", "The document delimiter." ),
						new FlaggedOption( "factory", MG4JClassParser.getParser(), IdentityDocumentFactory.class.getName(), JSAP.NOT_REQUIRED, 'f', "factory",
								"A document factory with a standard constructor." ),
						new FlaggedOption( "property", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'p', "property", "A 'key=value' specification, or the name of a property file." )
								.setAllowMultipleDeclarations( true ),
						new FlaggedOption( "indexedField", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, 'I', "indexed-field", "The field of the document factory that will be indexed." ),
						new FlaggedOption( "termProcessor", new ObjectParser( TermProcessor.class, MG4JClassParser.PACKAGE, new String[] { "getInstance" } ), NullTermProcessor.class.getName(),
								JSAP.NOT_REQUIRED, 't', "termprocessor", "The term pocessor to be used." ),
						new UnflaggedOption( "basename", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The archive output basename." ) } );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		if ( jsapResult.userSpecified( "sequence" ) && jsapResult.userSpecified( "objectSequence" ) ) throw new IllegalArgumentException(
				"You cannot specify both a serialised and an parseable-object sequence" );

		final DocumentSequence sequence = jsapResult.userSpecified( "objectSequence" ) ? (DocumentSequence)jsapResult.getObject( "objectSequence" ) : Scan.getSequence( jsapResult
				.getString( "sequence" ), jsapResult.getClass( "factory" ), jsapResult.getStringArray( "property" ), jsapResult.getInt( "delimiter" ), LOGGER );

		termProcessor = (TermProcessor)jsapResult.getObject( "termProcessor" );

		run( jsapResult.getString( "basename" ), sequence, termProcessor, jsapResult.getString( "indexedField" ) );

		sequence.close();
	}
}
